import os
import copy
import re
import functools
import subprocess
from threading import Thread
from Queue import Queue, Empty
import json

from populus import utils


is_geth_available = functools.partial(utils.is_executable_available, 'geth')
is_nice_available = functools.partial(utils.is_executable_available, 'nice')


POPULUS_DIR = os.path.abspath(os.path.dirname(__file__))


def get_blockchains_dir(project_dir):
    blockchains_dir = os.path.abspath(os.path.join(project_dir, 'chains'))
    utils.ensure_path_exists(blockchains_dir)
    return blockchains_dir


def get_geth_data_dir(project_dir, name):
    blockchains_dir = get_blockchains_dir(project_dir)
    data_dir = os.path.join(blockchains_dir, name)
    utils.ensure_path_exists(data_dir)
    return data_dir


def enqueue_stream(stream, queue):
    """
    Synchronously reads the output from a stream (stdout, stderr) and puts it
    onto the queue to be read.
    """
    for line in iter(stream.readline, b''):
        queue.put(line)


class PopenWrapper(subprocess.Popen):
    _locked = False

    def __init__(self, args, bufsize='1', stdin=subprocess.PIPE,
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs):
        super(PopenWrapper, self).__init__(args, bufsize=bufsize, stdin=stdin,
                                           stdout=stdout, stderr=stderr,
                                           **kwargs)
        self.stdout_queue = Queue()
        self.stdout_thread = Thread(
            target=enqueue_stream,
            args=(self.stdout, self.stdout_queue),
        )
        self.stdout_thread.daemon = True
        self.stdout_thread.start()

        self.stderr_queue = Queue()
        self.stderr_thread = Thread(
            target=enqueue_stream,
            args=(self.stderr, self.stderr_queue),
        )
        self.stderr_thread.daemon = True
        self.stderr_thread.start()

    def get_stdout_nowait(self):
        try:
            return self.stdout_queue.get_nowait()
        except Empty:
            return None

    def get_stderr_nowait(self):
        try:
            return self.stderr_queue.get_nowait()
        except Empty:
            return None

    _output_generator = None

    def get_output_nowait(self):
        if self._output_generator is None:
            def output_generator():
                while True:
                    yield self.get_stdout_nowait()
                    yield self.get_stderr_nowait()
            self._output_generator = output_generator()
        return self._output_generator.next()

    def communicate(self, *args, **kwargs):
        raise ValueError("Cannot communicate with a PopenWrapper")


DEFAULT_PW_PATH = os.path.join(POPULUS_DIR, 'default_blockchain_password')


def geth_wrapper(data_dir, popen_class=subprocess.Popen, cmd="geth",
                 genesis_block=None, miner_threads='1', extra_args=None,
                 max_peers='0', network_id='123456', no_discover=True,
                 mine=False, nice=True, unlock='0', password=DEFAULT_PW_PATH,
                 port=None):
    if nice and is_nice_available():
        command = ['nice', '-n', '20', cmd]
    else:
        command = [cmd]

    if genesis_block is None:
        genesis_block = get_genesis_block_path(data_dir)
        if not os.path.exists(genesis_block):
            genesis_block = os.path.join(POPULUS_DIR, 'genesis-test.json')

    if port is None:
        port = utils.get_open_port()

    command.extend((
        '--genesis', genesis_block,
        '--datadir', data_dir,
        '--maxpeers', max_peers,
        '--networkid', network_id,
        '--port', port,
    ))

    if miner_threads is not None:
        command.extend(('--minerthreads', miner_threads))

    if unlock is not None:
        command.extend((
            '--unlock', unlock,
        ))

    if password is not None:
        command.extend((
            '--password', password,
        ))

    if no_discover:
        command.append('--nodiscover')

    if mine:
        if unlock is None:
            raise ValueError("Cannot mine without an unlocked account")
        command.append('--mine')

    if extra_args:
        command.extend(extra_args)

    proc = popen_class(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=1,
    )
    return command, proc


def get_geth_accounts(data_dir, **kwargs):
    """
    Returns all geth accounts as tuple of hex encoded strings

    >>> geth_accounts()
    ... ('0x...', '0x...')
    """
    command, proc = geth_wrapper(data_dir, extra_args=['account', 'list'], **kwargs)
    stdoutdata, stderrdata = proc.communicate()

    if proc.returncode:
        if "no keys in store" in stderrdata:
            return tuple()
        else:
            raise subprocess.CalledProcessError(1, ' '.join(command))
    accounts = parse_geth_accounts(stdoutdata)
    return accounts


def create_geth_account(data_dir, **kwargs):
    command, proc = geth_wrapper(data_dir, extra_args=['account', 'new'], **kwargs)
    stdoutdata, stderrdata = proc.communicate()

    if proc.returncode:
        raise subprocess.CalledProcessError(1, ' '.join(command))

    match = account_regex.search(stdoutdata)
    if not match:
        raise ValueError("No address found in output: '{0}'".format(stdoutdata))

    return '0x' + match.groups()[0]


default_genesis_data = {
    "nonce": "0xdeadbeefdeadbeef",
    "timestamp": "0x0",
    "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
    "extraData": "0x686f727365",
    "gasLimit": "0x80000000000000",
    "difficulty": "0x400",
    "mixhash": "0x0000000000000000000000000000000000000000000000000000000000000000",
    "coinbase": "0x3333333333333333333333333333333333333333",
    "alloc": {}
}


def get_genesis_block_path(data_dir):
    return os.path.join(data_dir, 'genesis-block.json')


def ensure_account_exists(data_dir):
    accounts = get_geth_accounts(data_dir)
    if not accounts:
        account = create_geth_account(data_dir)
        genesis_block_path = get_genesis_block_path(data_dir)
        if not os.path.exists(genesis_block_path):
            genesis_data = copy.deepcopy(default_genesis_data)
            genesis_data['alloc'][account] = {
                "balance": "1000000000000000000000000000",  # 1,000,000,000 ether
            }
            with open(genesis_block_path, 'w') as genesis_block_file:
                genesis_block_file.write(json.dumps(genesis_data))
    else:
        account = accounts[0]
    return account


account_regex = re.compile('\{([a-f0-9]{40})\}')


def parse_geth_accounts(raw_accounts_output):
    accounts = account_regex.findall(raw_accounts_output)
    return tuple('0x' + account for account in accounts)


def run_geth_node(data_dir, rpc_server=True, rpc_addr=None, rpc_port=None,
                  mine=True, **kwargs):
    extra_args = []

    if rpc_server:
        extra_args.append('--rpc')

        if rpc_addr is not None:
            extra_args.extend(('--rpcaddr', rpc_addr))

        if rpc_port is not None:
            extra_args.extend(('--rpcport', rpc_port))

    command, proc = geth_wrapper(data_dir, popen_class=PopenWrapper,
                                 extra_args=extra_args, mine=mine, **kwargs)
    return command, proc


def reset_chain(data_dir):
    blockchain_dir = os.path.join(data_dir, 'chaindata')
    utils.remove_dir_if_exists(blockchain_dir)

    dapp_dir = os.path.join(data_dir, 'dapp')
    utils.remove_dir_if_exists(dapp_dir)

    nodekey_path = os.path.join(data_dir, 'nodekey')
    utils.remove_file_if_exists(nodekey_path)

    geth_ipc_path = os.path.join(data_dir, 'geth.ipc')
    utils.remove_file_if_exists(geth_ipc_path)



import binascii
import copy
import hashlib
import itertools

from ethereum import utils as ethereum_utils
from ethereum import abi

from populus.utils import wait_for_transaction


def decode_single(typ, data):
    base, sub, _ = abi.process_type(typ)

    if base == 'address':
        return '0x' + data[-40:]
    elif base == 'string' or base == 'bytes' or base == 'hash':
        if sub:
            bytes = ethereum_utils.int_to_32bytearray(int(data, 16))
            while bytes and bytes[-1] == 0:
                bytes.pop()
            if bytes:
                return ''.join(chr(b) for b in bytes)
        else:
            num_bytes = int(data[64 + 2:128 + 2], 16)
            bytes_as_hex = data[2 + 128:2 + 128 + (2 * num_bytes)]
            return ethereum_utils.decode_hex(bytes_as_hex)
    elif base == 'uint':
        return int(data, 16)
    elif base == 'int':
        o = int(data, 16)
        return (o - 2 ** int(sub)) if o >= 2 ** (int(sub) - 1) else o
    elif base == 'ureal':
        raise NotImplementedError('havent gotten to this')
        high, low = [int(x) for x in sub.split('x')]
        # return big_endian_to_int(data) * 1.0 / 2 ** low
    elif base == 'real':
        raise NotImplementedError('havent gotten to this')
        high, low = [int(x) for x in sub.split('x')]
        # return (big_endian_to_int(data) * 1.0 / 2 ** low) % 2 ** high
    elif base == 'bool':
        return bool(int(data, 16))
    else:
        raise ValueError("Unknown base: `{0}`".format(base))


def decode_multi(types, outputs):
    res = abi.decode_abi(types, binascii.a2b_hex(outputs[2:]))
    return res


def strip_0x_prefix(value):
    if value.startswith('0x'):
        return value[2:]
    return value


def clean_args(*args):
    for _type, arg in args:
        if _type == 'address':
            yield strip_0x_prefix(arg)
        else:
            yield arg


class ContractBound(object):
    _contract = None

    def _bind(self, contract):
        self._contract = contract

    @property
    def contract(self):
        if self._contract is None:
            raise AttributeError("Function not bound to a contract")
        return self._contract


class Function(ContractBound):
    def __init__(self, name, inputs=None, outputs=None, constant=False):
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.constant = constant

    def __str__(self):
        signature = "{func_name}({arg_types})".format(
            func_name=self.name,
            arg_types=', '.join(
                "{0} {1}".format(i['type'], i['name']) for i in self.inputs
            )
        )
        return signature

    def __copy__(self):
        return self.__class__(self.name, self.inputs, self.outputs, self.constant)

    @property
    def input_types(self):
        """
        Iterable of the types this function takes.
        """
        if self.inputs:
            return [i['type'] for i in self.inputs]
        return []

    @property
    def output_types(self):
        """
        Iterable of the types this function takes.
        """
        if self.outputs:
            return [i['type'] for i in self.outputs]
        return []

    @property
    def abi_function_signature(self):
        """
        Compute the bytes4 signature for the function.
        """
        signature = "{func_name}({arg_types})".format(
            func_name=self.name,
            arg_types=','.join(self.input_types),
        )
        return ethereum_utils.big_endian_to_int(ethereum_utils.sha3(signature)[:4])

    @property
    def encoded_abi_function_signature(self):
        return ethereum_utils.zpad(ethereum_utils.encode_int(self.abi_function_signature), 4)

    def abi_args_signature(self, args):
        """
        Given the calling `args` for the function call, abi encode them.
        """
        if len(self.input_types) != len(args):
            raise ValueError("Expected {0} arguments, only got {1}".format(len(self.input_types), len(args)))  # NOQA
        scrubbed_args = tuple(clean_args(*zip(self.input_types, args)))
        return abi.encode_abi(self.input_types, scrubbed_args)

    def get_call_data(self, args):
        """
        TODO: this needs tests.
        """
        prefix = self.encoded_abi_function_signature
        suffix = self.abi_args_signature(args)
        data = "{0}{1}".format(prefix, suffix)
        return ethereum_utils.encode_hex(data)

    def cast_return_data(self, outputs):
        if len(self.output_types) != 1:
            return decode_multi(self.output_types, outputs)
        output_type = self.output_types[0]

        return decode_single(output_type, outputs)

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        else:
            return obj._meta.functions[self.name]

    def __call__(self, *args, **kwargs):
        if self.constant:
            return self.call(*args, **kwargs)
        return self.sendTransaction(*args, **kwargs)

    def sendTransaction(self, *args, **kwargs):
        data = self.get_call_data(args)

        if 'gas' not in kwargs:
            kwargs['gas'] = get_max_gas(self.contract._meta.rpc_client)

        return self.contract._meta.rpc_client.send_transaction(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )

    def call(self, *args, **kwargs):
        raw = kwargs.pop('raw', False)
        data = self.get_call_data(args)

        output = self.contract._meta.rpc_client.call(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )
        if raw:
            return output
        return self.cast_return_data(output)


class Event(ContractBound):
    """
    {
        'inputs': [
            {'indexed': False, 'type': 'bytes32', 'name': 'dataHash'},
            {'indexed': False, 'type': 'bytes', 'name': 'data'},
        ],
        'type': 'event',
        'name': 'DataRegistered',
        'anonymous': False,
    }
    """
    def __init__(self, name, inputs, anonymous):
        self.name = name
        self.inputs = inputs
        self.anonymous = anonymous

    def __call__(self, *args):
        pass

    def __copy__(self):
        return self.__class__(self.name, self.inputs, self.anonymous)


class ContractBase(object):
    def __init__(self, address, rpc_client):
        functions = {fn.name: fn for fn in (copy.copy(f) for f in self._config._functions)}
        events = {ev.name: ev for ev in (copy.copy(e) for e in self._config._events)}
        for obj in itertools.chain(functions.values(), events.values()):
            obj._bind(self)
        self._meta = ContractMeta(address, rpc_client, functions, events)

    def __str__(self):
        return "{name}({address})".format(name=self.__class__.__name__, address=self.address)

    @classmethod
    def get_deploy_data(cls, *args):
        data = cls._config.code
        if args:
            if cls._config.constructor is None:
                raise ValueError("This contract does not appear to have a constructor")
            data += ethereum_utils.encode_hex(cls._config.constructor.abi_args_signature(args))

        return data

    #
    #  Instance Methods
    #
    def get_balance(self, block="latest"):
        return self._meta.rpc_client.get_balance(self._meta.address, block=block)


class ContractMeta(object):
    """
    Instance level contract data.
    """
    def __init__(self, address, rpc_client, functions, events):
        self.address = address
        self.rpc_client = rpc_client
        self.functions = functions
        self.events = events


class Config(object):
    """
    Contract (class) level contract data.
    """
    def __init__(self, code, source, abi, functions, events, constructor):
        self.code = code
        self.source = source
        self.abi = abi
        self._functions = functions
        self._events = events
        self.constructor = constructor


def Contract(contract_meta, contract_name=None):
    _abi = contract_meta['info']['abiDefinition']
    code = contract_meta['code']
    source = contract_meta['info']['source']

    if contract_name is None:
        contract_name = "Unknown-{0}".format(hashlib.md5(code).hexdigest())

    functions = []
    events = []
    constructor = None

    _dict = {}

    for signature_item in _abi:
        if signature_item['type'] == 'constructor':
            # Constructors don't need to be part of a contract's methods
            if signature_item.get('inputs'):
                constructor = Function(
                    name='constructor',
                    inputs=signature_item['inputs'],
                )
            continue

        if signature_item['name'] in _dict:
            # TODO: handle namespace conflicts
            raise ValueError("About to overwrite a function signature for duplicate function name {0}".format(signature_item['name']))  # NOQA

        if signature_item['type'] == 'function':
            # make sure we're not overwriting a signature

            func = Function(
                name=signature_item['name'],
                inputs=signature_item['inputs'],
                outputs=signature_item['outputs'],
                constant=signature_item['constant'],
            )
            _dict[signature_item['name']] = func
            functions.append(func)
        elif signature_item['type'] == 'event':
            event = Event(
                name=signature_item['name'],
                inputs=signature_item['inputs'],
                anonymous=signature_item['anonymous'],
            )
            _dict[signature_item['name']] = event
            events.append(event)
        else:
            raise ValueError("Unknown signature item '{0}'".format(signature_item))

    docstring = """
    contract {contract_name} {{
    // Events
    {events}

    // Functions
    {functions}
    }}
    """.format(
        contract_name=contract_name,
        functions='\n'.join(str(f) for f in functions),
        events='\n'.join(str(e) for e in events),
    )

    _dict['__doc__'] = docstring
    _dict['_config'] = Config(code, source, _abi, functions, events, constructor)

    return type(str(contract_name), (ContractBase,), _dict)


def deploy_contract(rpc_client, contract_class, constructor_args=None, **kwargs):
    if 'data' in kwargs:
        raise ValueError("Cannot supply `data` for contract deployment")

    if constructor_args is None:
        constructor_args = []

    kwargs['data'] = contract_class.get_deploy_data(*constructor_args)
    txn_hash = rpc_client.send_transaction(**kwargs)
    return txn_hash


def get_contract_address_from_txn(rpc_client, txn_hash, max_wait=0):
    txn_receipt = wait_for_transaction(rpc_client, txn_hash, max_wait)

    return txn_receipt['contractAddress']


def get_max_gas(rpc_client, scale=0.95):
    latest_block = rpc_client.get_block_by_number('latest')
    max_gas_hex = latest_block['gasLimit']
    max_gas = int(max_gas_hex, 16)
    return int(max_gas * scale)



import os
import time
import signal
import multiprocessing

import pytest

import click

from eth_rpc_client import Client

from populus import utils
from populus.compilation import (
    get_contracts_dir,
    compile_and_write_contracts,
)
from populus.observers import (
    get_contracts_observer,
    get_contracts_code_observer,
    get_static_assets_observer,
)
from populus.web import (
    get_flask_app,
    collect_static_assets,
    compile_js_contracts,
    get_html_dir,
    write_default_index_html_document,
    get_static_assets_dir,
)
from populus.geth import (
    get_geth_data_dir,
    run_geth_node,
    ensure_account_exists,
    reset_chain,
)


@click.group()
def main():
    pass


@main.command()
def init():
    """
    Generate project layout with an example contract.
    """
    project_dir = os.getcwd()
    contracts_dir = get_contracts_dir(project_dir)
    if utils.ensure_path_exists(contracts_dir):
        click.echo("Created Directory: ./{0}".format(os.path.relpath(contracts_dir)))

    example_contract_path = os.path.join(contracts_dir, 'Example.sol')
    if not os.path.exists(example_contract_path):
        with open(example_contract_path, 'w') as example_contract_file:
            example_contract_file.write('contract Example {\n        function Example() {\n        }\n}\n')  # NOQA
        click.echo("Created Example Contract: ./{0}".format(os.path.relpath(example_contract_path)))  # NOQA

    tests_dir = os.path.join(project_dir, 'tests')
    if utils.ensure_path_exists(tests_dir):
        click.echo("Created Directory: ./{0}".format(os.path.relpath(tests_dir)))

    example_tests_path = os.path.join(tests_dir, 'test_example.py')
    if not os.path.exists(example_tests_path):
        with open(example_tests_path, 'w') as example_tests_file:
            example_tests_file.write('def test_it(deployed_contracts):\n    example = deployed_contracts.Example\n    assert example.address\n')  # NOQA
        click.echo("Created Example Tests: ./{0}".format(os.path.relpath(example_tests_path)))  # NOQA

    initialize_web_directories(project_dir)


@main.command('compile')
@click.option(
    '--watch',
    '-w',
    is_flag=True,
    help="Watch contract source files and recompile on changes",
)
@click.argument('filters', nargs=-1)
def compile_contracts(watch, filters):
    """
    Compile project contracts, storing their output in `./build/contracts.json`

    Call bare to compile all contracts or specify contract names or file paths
    to restrict to only compiling those contracts.

    Pass in a file path and a contract name separated by a colon(":") to
    specify only named contracts in the specified file.
    """
    project_dir = os.getcwd()

    click.echo("============ Compiling ==============")
    click.echo("> Loading contracts from: {0}".format(get_contracts_dir(project_dir)))

    result = compile_and_write_contracts(project_dir, *filters)
    contract_source_paths, compiled_sources, output_file_path = result

    click.echo("> Found {0} contract source files".format(len(contract_source_paths)))
    for path in contract_source_paths:
        click.echo("- {0}".format(os.path.basename(path)))
    click.echo("")
    click.echo("> Compiled {0} contracts".format(len(compiled_sources)))
    for contract_name in sorted(compiled_sources.keys()):
        click.echo("- {0}".format(contract_name))
    click.echo("")
    click.echo("> Outfile: {0}".format(output_file_path))

    if watch:
        # The path to watch
        click.echo("============ Watching ==============")

        observer = get_contracts_observer(project_dir, filters)
        observer.start()
        try:
            while observer.is_alive():
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()


@main.command()
def deploy():
    """
    Deploy contracts.
    """
    contracts = utils.load_contracts(os.getcwd())
    client = Client('127.0.0.1', '8545')

    deployed_contracts = utils.deploy_contracts(client, contracts)

    name_padding = max(len(n) + 1 for n in deployed_contracts.keys())
    for name, info in deployed_contracts.items():
        click.echo("{name} @ {addr} via txn:{txn_hash}".format(
            name=name.ljust(name_padding),
            addr=(info.get('addr') or '<pending>').ljust(42),
            txn_hash=info['txn'].ljust(66),
        ))


@main.command()
def test():
    """
    Test contracts (wrapper around py.test)
    """
    pytest.main(os.path.join(os.getcwd(), 'tests'))


@main.group()
def web():
    """
    HTML/CSS/JS tooling.
    """
    pass


def initialize_web_directories(project_dir):
    html_dir = get_html_dir(project_dir)
    if utils.ensure_path_exists(html_dir):
        click.echo("Created Directory: ./{0}".format(os.path.relpath(html_dir)))

    html_index_path = os.path.join(html_dir, 'index.html')
    if not os.path.exists(html_index_path):
        write_default_index_html_document(html_index_path)
        click.echo("Created HTML Index File: ./{0}".format(os.path.relpath(html_index_path)))

    static_assets_dir = get_static_assets_dir(project_dir)
    if utils.ensure_path_exists(static_assets_dir):
        click.echo("Created Directory: ./{0}".format(os.path.relpath(static_assets_dir)))


@web.command('init')
def web_init():
    project_dir = os.getcwd()
    initialize_web_directories(project_dir)


@web.command('runserver')
@click.option('--debug/--no-debug', default=True)
def web_runserver(debug):
    """
    Run the development server.
    """
    project_dir = os.getcwd()
    # Do initial setup
    click.echo("Compiling contracts...")
    compile_and_write_contracts(project_dir)
    click.echo("Compiling contracts.js...")
    compile_js_contracts(project_dir)
    click.echo("Collectind static assets...")
    collect_static_assets(project_dir)

    all_threads = []

    # Contract Builder Thread
    contracts_observer_thread = get_contracts_observer(project_dir)
    contracts_observer_thread.daemon = True

    # Contract JS Builder Thread
    contracts_code_observer_thread = get_contracts_code_observer(project_dir)
    contracts_code_observer_thread.daemon = True

    # Assets Collector Thread
    static_assets_observer_thread = get_static_assets_observer(project_dir)
    static_assets_observer_thread.daemon = True

    # Webserver Thread
    flask_app = get_flask_app(project_dir)
    webserver_thread = multiprocessing.Process(
        target=flask_app.run,
        kwargs={'use_reloader': False, 'debug': debug},
    )
    webserver_thread.daemon = True

    # Start all the threads
    contracts_observer_thread.start()
    contracts_code_observer_thread.start()
    static_assets_observer_thread.start()
    webserver_thread.start()

    try:
        all_threads = (
            contracts_observer_thread,
            contracts_code_observer_thread,
            static_assets_observer_thread,
            webserver_thread,
        )
        while any(t.is_alive() for t in all_threads):
            if not all(t.is_alive() for t in all_threads):
                raise click.Abort("Some threads died!")
            time.sleep(1)
    except KeyboardInterrupt:
        for t in all_threads:
            if hasattr(t, 'stop'):
                t.stop()
            elif hasattr(t, 'terminate'):
                t.terminate()
            else:
                raise ValueError("wat")
        for t in all_threads:
            t.join()


@web.command('collect')
def web_collect():
    """
    Compile the contracts.js file and collect static assets.
    """
    compile_js_contracts(os.getcwd())
    collect_static_assets(os.getcwd())


@main.group()
def chain():
    """
    Wrapper around `geth`.
    """
    pass


@chain.command('reset')
@click.argument('name', nargs=1, default="default")
@click.option('--confirm/--no-confirm', default=True)
def chain_reset(name, confirm):
    """
    Reset a test chain
    """
    data_dir = get_geth_data_dir(os.getcwd(), name)
    if confirm and not click.confirm("Are you sure you want to reset blockchain '{0}': {1}".format(name, data_dir)):  # NOQA
        raise click.Abort()
    reset_chain(data_dir)


@chain.command('run')
@click.argument('name', nargs=1, default="default")
@click.option('--mine/--no-mine', default=True)
def chain_run(name, mine):
    """
    Run a geth node.
    """
    data_dir = get_geth_data_dir(os.getcwd(), name)

    if not os.path.exists(data_dir):
        if name == 'default':
            utils.ensure_path_exists(data_dir)
        elif click.confirm("The chain '{0}' does not exist.  Would you like to create it?".format(name)):  # NOQA
            utils.ensure_path_exists(data_dir)
        else:
            raise click.UsageError("Unknown chain '{0}'".format(name))

    ensure_account_exists(data_dir)

    command, proc = run_geth_node(data_dir, mine=mine)

    click.echo("Running: '{0}'".format(' '.join(command)))

    try:
        while True:
            out_line = proc.get_stdout_nowait()
            if out_line:
                click.echo(out_line, nl=False)

            err_line = proc.get_stderr_nowait()
            if err_line:
                click.echo(err_line, nl=False)

            if err_line is None and out_line is None:
                time.sleep(0.2)
    except KeyboardInterrupt:
        try:
            proc.send_signal(signal.SIGINT)
            # Give the subprocess a SIGINT and give it a few seconds to
            # cleanup.
            utils.wait_for_popen(proc)
            while not proc.stdout_queue.empty() or not proc.stderr_queue.empty():
                out_line = proc.get_stdout_nowait()
                if out_line:
                    click.echo(out_line, nl=False)

                err_line = proc.get_stderr_nowait()
                if err_line:
                    click.echo(err_line, nl=False)
        except:
            # Try a harder termination.
            proc.terminate()
            utils.wait_for_popen(proc, 2)
    if proc.poll() is None:
        # Force it to kill if it hasn't exited already.
        proc.kill()
    if proc.returncode:
        raise click.ClickException("Error shutting down geth process.")



import os

import click

from watchdog.observers.polling import (
    PollingObserver,
)
from watchdog.events import FileSystemEventHandler

from populus import utils
from populus.web import (
    get_contracts_js_path,
    get_static_assets_dir,
    compile_js_contracts,
    collect_static_assets,
    project_has_assets,
)
from populus.compilation import (
    compile_and_write_contracts,
    get_compiled_contract_destination_path,
)


class ContractSourceChangedEventHandler(FileSystemEventHandler):
    """
    > http://pythonhosted.org/watchdog/api.html#watchdog.events.FileSystemEventHandler
    """
    def __init__(self, *args, **kwargs):
        self.project_dir = kwargs.pop('project_dir')
        self.contract_filters = kwargs.pop('contract_filters')

    def on_any_event(self, event):
        click.echo("============ Detected Change ==============")
        click.echo("> {0} => {1}".format(event.event_type, event.src_path))
        click.echo("> recompiling...")
        compile_and_write_contracts(self.project_dir, *self.contract_filters)
        click.echo("> watching...")


def get_contracts_observer(project_dir, contract_filters=None):
    if contract_filters is None:
        contract_filters = []
    watch_path = utils.get_contracts_dir(project_dir)

    event_handler = ContractSourceChangedEventHandler(
        project_dir=project_dir,
        contract_filters=contract_filters,
    )
    observer = PollingObserver()
    observer.schedule(event_handler, watch_path, recursive=True)
    return observer


class ContractCodeChangedEventHandler(FileSystemEventHandler):
    def __init__(self, *args, **kwargs):
        self.project_dir = kwargs.pop('project_dir')

    def on_any_event(self, event):
        is_contracts_json_file = os.path.samefile(
            event.src_path,
            get_compiled_contract_destination_path(self.project_dir),
        )
        if not is_contracts_json_file:
            return

        click.echo("============ Detected Change ==============")
        click.echo("> {0} => {1}".format(event.event_type, event.src_path))
        click.echo("> Recreating 'contracts.js' ...")
        compile_js_contracts(self.project_dir)
        click.echo("> watching...")


def get_contracts_code_observer(project_dir):
    observer = PollingObserver()

    build_path = utils.get_build_dir(project_dir)
    event_handler = ContractCodeChangedEventHandler(project_dir=project_dir)
    observer.schedule(event_handler, build_path, recursive=False)

    return observer


class ContractsJSChangedEventHandler(FileSystemEventHandler):
    def __init__(self, *args, **kwargs):
        self.project_dir = kwargs.pop('project_dir')

    def on_any_event(self, event):
        is_contracts_js_file = os.path.samefile(
            event.src_path,
            get_contracts_js_path(self.project_dir),
        )
        if not is_contracts_js_file:
            click.echo("Ignoring: {0}".format(event.src_path))
            return

        click.echo("============ Detected Change ==============")
        click.echo("> {0} => {1}".format(event.event_type, event.src_path))
        click.echo("> Collecting Assets...")
        collect_static_assets(self.project_dir)
        click.echo("> watching...")


class AssetsChangedEventHandler(FileSystemEventHandler):
    def __init__(self, *args, **kwargs):
        self.project_dir = kwargs.pop('project_dir')

    def on_any_event(self, event):
        click.echo("============ Detected Change ==============")
        click.echo("> {0} => {1}".format(event.event_type, event.src_path))
        click.echo("> Collecting Assets...")
        collect_static_assets(self.project_dir)
        click.echo("> watching...")


def get_static_assets_observer(project_dir):
    build_path = utils.get_build_dir(project_dir)

    observer = PollingObserver()

    contracts_js_event_handler = ContractsJSChangedEventHandler(project_dir=project_dir)
    observer.schedule(contracts_js_event_handler, build_path, recursive=False)

    if project_has_assets(project_dir):
        assets_watch_path = get_static_assets_dir(project_dir)
        assets_event_handler = AssetsChangedEventHandler(project_dir=project_dir)
        observer.schedule(assets_event_handler, assets_watch_path, recursive=True)

    return observer



import itertools
import glob
import os
import json
import functools

from populus.utils import (
    get_contracts_dir,
    get_build_dir,
)
from populus.solidity import solc


def find_project_contracts(project_dir):
    contracts_dir = get_contracts_dir(project_dir)
    # TODO: support non-solidity based contract compilation.
    solidity_glob = os.path.join(contracts_dir, "*.sol")
    serpent_glob = os.path.join(contracts_dir, "*.se")
    lll_glob = os.path.join(contracts_dir, "*.lll")
    mutan_glob = os.path.join(contracts_dir, "*.mutan")

    return tuple(itertools.chain(
        glob.glob(solidity_glob),
        glob.glob(serpent_glob),
        glob.glob(lll_glob),
        glob.glob(mutan_glob),
    ))


def get_compiled_contract_destination_path(project_dir):
    build_dir = get_build_dir(project_dir)
    file_path = os.path.join(build_dir, 'contracts.json')
    return file_path


def write_compiled_sources(project_dir, compiled_sources):
    file_path = get_compiled_contract_destination_path(project_dir)
    with open(file_path, 'w') as outfile:
        outfile.write(
            json.dumps(compiled_sources, sort_keys=True, indent=4, separators=(',', ': '))
        )
    return file_path


def get_compiler_for_file(file_path):
    _, _, ext = file_path.rpartition('.')

    if ext == 'sol':
        return solc
    elif ext == 'lll':
        raise ValueError("Compilation of LLL contracts is not yet supported")
    elif ext == 'mu':
        raise ValueError("Compilation of LLL contracts is not yet supported")
    elif ext == 'se':
        raise ValueError("Compilation of LLL contracts is not yet supported")

    raise ValueError("Unknown contract extension {0}".format(ext))


def compile_source_file(source_path):
    compiler = get_compiler_for_file(source_path)

    with open(source_path) as source_file:
        source_code = source_file.read()

    # TODO: solidity specific
    compiled_source = compiler(source_code)
    return compiled_source


def compile_project_contracts(contracts_dir, filters=None):
    compiled_sources = {}

    for source_path in contracts_dir:
        compiled_source = compile_source_file(source_path)

        if filters:
            for contract_name, contract_data in compiled_source.items():
                if any(f(source_path, contract_name) for f in filters):
                    compiled_sources[contract_name] = contract_data
        else:
            compiled_sources.update(compiled_source)

    return compiled_sources


def check_if_matches_filter(file_path_filter, contract_filter, file_path, contract_name):
    if file_path_filter == contract_filter:
        allow_either = True
    else:
        allow_either = False

    file_path_match = all((
        file_path.endswith(file_path_filter),  # Same path
        os.path.basename(file_path_filter) == os.path.basename(file_path),  # same filename
    ))
    name_match = contract_filter == contract_name

    if file_path_match and name_match:
        return True
    elif allow_either and (file_path_match or name_match):
        return True
    else:
        return False


def generate_filter(filter_text):
    """
    Takes one of the following formats.

    * `ContractName`
    * `path/to/contractFile.sol`
    * `path/to/contractFile.sol:ContractName`

    and Returns callables that return `True` if the contract should be included.
    """
    if ':' in filter_text:
        file_path_filter, _, contract_filter = filter_text.partition(':')
    else:
        file_path_filter = contract_filter = filter_text

    return functools.partial(check_if_matches_filter, file_path_filter, contract_filter)


def get_contract_filters(*contracts):
    """
    Generate the filter functions for contract compilation.
    """
    return [generate_filter(filter_text) for filter_text in contracts]


def compile_and_write_contracts(project_dir, *contracts):
    filters = get_contract_filters(*contracts)
    contract_source_paths = find_project_contracts(project_dir)

    compiled_sources = compile_project_contracts(contract_source_paths, filters)

    output_file_path = write_compiled_sources(project_dir, compiled_sources)
    return contract_source_paths, compiled_sources, output_file_path



import os
import shutil

from flask import (
    Flask,
    current_app,
)

from populus.utils import (
    get_build_dir,
    ensure_path_exists,
)


POPULUS_ASSET_PATH = os.path.join(os.path.dirname(__file__), 'assets')
POPULUS_TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')


def get_build_assets_dir(project_dir):
    build_dir = get_build_dir(project_dir)
    return os.path.abspath(os.path.join(build_dir, 'assets'))


def get_html_dir(project_dir):
    html_dir = os.path.abspath(os.path.join(project_dir, './html'))
    return html_dir


def index_view():
    html_dir = get_html_dir(current_app.populus_project_dir)
    document_path = os.path.join(html_dir, 'index.html')
    with open(document_path) as document_file:
        document = document_file.read()
    return document


def write_default_index_html_document(dest_path):
    src_path = os.path.join(POPULUS_TEMPLATES_PATH, 'index.html')
    shutil.copy(src_path, dest_path)


def get_flask_app(project_dir):
    static_folder = os.path.join(project_dir, 'build', 'assets')
    app = Flask(__name__, static_folder=static_folder, static_url_path='/static')
    app.populus_project_dir = project_dir

    app.route('/')(index_view)
    return app


js_contracts_template = """
var contracts = contracts || {{}};

function makeContracts() {{
    var contractData = {contract_data};
    var contractNames = Object.keys(contractData);
    for (var i=0; i < contractNames.length; i++) {{
        var contractName = contractNames[i];
        contracts[contractName] = web3.eth.contract(contractData[contractName].info.abiDefinition);
    }}
}};
makeContracts();
"""


def get_contracts_js_path(project_dir):
    build_dir = get_build_dir(project_dir)
    return os.path.join(build_dir, 'contracts.js')


def compile_js_contracts(project_dir):
    build_dir = get_build_dir(project_dir)

    contracts_json_path = os.path.join(build_dir, 'contracts.json')
    with open(contracts_json_path) as contracts_json_file:
        contracts_json = contracts_json_file.read()

    js_contracts = js_contracts_template.format(contract_data=contracts_json)

    js_contracts_path = get_contracts_js_path(project_dir)

    with open(js_contracts_path, 'w') as js_contracts_file:
        js_contracts_file.write(js_contracts)

    return js_contracts_path


def get_static_assets_dir(project_dir):
    return os.path.abspath(os.path.join(project_dir, 'assets'))


def project_has_assets(project_dir):
    project_assets_path = get_static_assets_dir(project_dir)
    return os.path.exists(project_assets_path)


def collect_static_assets(project_dir):
    project_assets_path = get_static_assets_dir(project_dir)

    build_dir = get_build_dir(project_dir)
    build_assets_path = os.path.join(build_dir, 'assets')
    ensure_path_exists(build_assets_path)
    build_js_assets_path = os.path.join(build_assets_path, 'js')
    ensure_path_exists(build_js_assets_path)

    # Put the contracts json in place.
    contracts_js_path = os.path.join(get_build_dir(project_dir), 'contracts.js')
    shutil.copy(contracts_js_path, build_js_assets_path)

    search_paths = (
        POPULUS_ASSET_PATH,
        project_assets_path,
    )

    for base_assets_dir in search_paths:
        prefix_length = len(base_assets_dir)
        for (dirpath, dirnames, filenames) in os.walk(base_assets_dir):
            asset_dir = os.path.join(build_assets_path, dirpath[prefix_length + 1:])
            ensure_path_exists(asset_dir)

            for filename in filenames:
                asset_path = os.path.abspath(os.path.join(dirpath, filename))
                shutil.copy(asset_path, asset_dir)



import os
import time
import threading

import pytest


@pytest.fixture(scope="session")
def test_coinbase():
    from ethereum import tester
    return tester.encode_hex(tester.accounts[0])


@pytest.yield_fixture()
def rpc_server(request):
    from testrpc.__main__ import create_server
    from testrpc.testrpc import evm_reset

    rpc_port = getattr(request.module, 'rpc_port', 8545)
    rpc_host = getattr(request.module, 'rpc_host', '127.0.0.1')

    server = create_server(rpc_host, rpc_port)

    evm_reset()

    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()

    yield server

    server.shutdown()
    server.server_close()


@pytest.fixture(scope="module")
def rpc_client(request):
    from eth_rpc_client import Client
    rpc_hostname = getattr(request.module, 'rpc_server_host', '127.0.0.1')
    rpc_port = getattr(request.module, 'rpc_server_port', '8545')
    client = Client(rpc_hostname, rpc_port)
    return client


@pytest.fixture(scope="module")
def contracts(request):
    from populus.utils import load_contracts
    from populus.contracts import Contract

    project_dir = getattr(request.module, "project_dir", os.getcwd())

    contracts = load_contracts(project_dir)

    contract_classes = {
        name: Contract(contract_meta, name) for name, contract_meta in contracts.items()
    }

    _dict = {
        '__iter__': lambda s: iter(contract_classes.items()),
        '__getitem__': lambda s, k: contract_classes.__getitem__[k],
    }
    _dict.update(contract_classes)

    return type('contracts', (object,), _dict)()


@pytest.fixture(scope="module")
def deployed_contracts(request, rpc_client, contracts):
    from populus.contracts import (
        deploy_contract,
        get_contract_address_from_txn,
        get_max_gas,
    )
    from populus.utils import (
        wait_for_block,
    )

    _dict = {}

    deploy_wait_for_block = getattr(request.module, 'deploy_wait_for_block', 0)
    deploy_wait_for_block_max_wait = getattr(request.module, 'deploy_wait_for_block_max_wait', 30)

    wait_for_block(rpc_client, deploy_wait_for_block, deploy_wait_for_block_max_wait)

    deploy_address = getattr(request.module, 'deploy_address', rpc_client.get_coinbase())
    deploy_max_wait = getattr(request.module, 'deploy_max_wait', 0)
    deploy_gas_limit = getattr(request.module, 'deploy_gas_limit', get_max_gas(rpc_client))

    for contract_name, contract_class in contracts:
        txn_hash = deploy_contract(
            rpc_client,
            contract_class,
            _from=deploy_address,
            gas=deploy_gas_limit,
        )
        contract_addr = get_contract_address_from_txn(
            rpc_client,
            txn_hash,
            max_wait=deploy_max_wait,
        )
        _dict[contract_name] = contract_class(contract_addr, rpc_client)

    return type('deployed_contracts', (object,), _dict)


@pytest.yield_fixture(scope="module")
def geth_node(request):
    from populus.geth import (
        run_geth_node,
        get_geth_data_dir,
        ensure_account_exists,
        reset_chain,
    )
    from populus.utils import (
        ensure_path_exists,
        kill_proc,
    )

    project_dir = getattr(request.module, 'geth_project_dir', os.getcwd())
    chain_name = getattr(request.module, 'geth_chain_name', 'default-test')
    data_dir = get_geth_data_dir(project_dir, chain_name)

    ensure_path_exists(data_dir)
    ensure_account_exists(data_dir)

    if getattr(request.module, 'geth_reset_chain', True):
        reset_chain(data_dir)

    rpc_port = getattr(request.module, 'rpc_port', '8545')
    rpc_host = getattr(request.module, 'rpc_host', '127.0.0.1')

    geth_max_wait = getattr(request.module, 'geth_max_wait', 5)

    command, proc = run_geth_node(data_dir, rpc_addr=rpc_host, rpc_port=rpc_port)

    start = time.time()
    while time.time() < start + geth_max_wait:
        output = []
        line = proc.get_output_nowait()
        if line:
            output.append(line)

        if line is None:
            continue
        if 'Starting mining operation' in line:
            break
        elif "Still generating DAG" in line:
            print(line[line.index("Still generating DAG"):])
        elif line.startswith('Fatal:'):
            kill_proc(proc)
            raise ValueError(
                "Geth Errored while starting\nerror: {0}\n\nFull Output{1}".format(
                    line, ''.join(output),
                )
            )
    else:
        kill_proc(proc)
        raise ValueError("Geth process never started\n\n{0}".format(''.join(output)))

    yield proc
    kill_proc(proc)


@pytest.fixture(scope='module')
def geth_coinbase(request):
    from populus.geth import (
        get_geth_data_dir,
        ensure_account_exists,
    )
    from populus.utils import (
        ensure_path_exists,
    )

    project_dir = getattr(request.module, 'project_dir', os.getcwd())
    chain_name = getattr(request.module, 'chain_name', 'default-test')
    data_dir = get_geth_data_dir(project_dir, chain_name)

    ensure_path_exists(data_dir)
    geth_coinbase = ensure_account_exists(data_dir)
    return geth_coinbase



import os
import shutil
import json
import socket
import time
import signal


CONTRACTS_DIR = "./contracts/"


def get_contracts_dir(project_dir):
    contracts_dir = os.path.join(project_dir, CONTRACTS_DIR)
    return os.path.abspath(contracts_dir)


BUILD_DIR = "./build/"


def ensure_path_exists(dir_path):
    """
    Make sure that a path exists
    """
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)
        return True
    return False


def remove_file_if_exists(path):
    if os.path.isfile(path):
        os.remove(path)
        return True
    return False


def remove_dir_if_exists(path):
    if os.path.isdir(path):
        shutil.rmtree(path)
        return True
    return False


def get_build_dir(project_dir):
    build_dir_path = os.path.join(project_dir, BUILD_DIR)
    ensure_path_exists(build_dir_path)
    return build_dir_path


def load_contracts(project_dir):
    compiled_contracts_path = os.path.join(project_dir, BUILD_DIR, 'contracts.json')
    if not os.path.exists(compiled_contracts_path):
        raise ValueError("No compiled contracts found")

    with open(compiled_contracts_path) as contracts_file:
        contracts = json.loads(contracts_file.read())

    return contracts


def deploy_contracts(client, contracts):
    _from = client.get_coinbase()
    deployed_contracts = {}

    for name, contract in contracts.items():
        txn_hash = client.send_transaction(_from=_from, data=contract['code'])
        receipt = client.get_transaction_receipt(txn_hash)
        if receipt is None:
            contract_addr = None
        else:
            contract_addr = receipt['contractAddress']
        deployed_contracts[name] = {
            'txn': txn_hash,
            'addr': contract_addr,
        }

    return deployed_contracts


def is_executable_available(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath = os.path.dirname(program)
    if fpath:
        if is_exe(program):
            return True
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return True

    return False


def get_open_port():
    sock = socket.socket()
    sock.bind(('127.0.0.1', 0))
    port = sock.getsockname()[1]
    sock.close()
    return str(port)


def wait_for_popen(proc, max_wait=5):
    wait_till = time.time() + 5
    while proc.poll() is None and time.time() < wait_till:
        time.sleep(0.1)


def kill_proc(proc):
    try:
        if proc.poll() is None:
            proc.send_signal(signal.SIGINT)
            wait_for_popen(proc, 5)
        if proc.poll() is None:
            proc.terminate()
            wait_for_popen(proc, 2)
        if proc.poll() is None:
            proc.kill()
            wait_for_popen(proc, 1)
    except KeyboardInterrupt:
        proc.kill()


def wait_for_transaction(rpc_client, txn_hash, max_wait=60):
    start = time.time()
    while True:
        txn_receipt = rpc_client.get_transaction_receipt(txn_hash)
        if txn_receipt is not None:
            break
        elif time.time() > start + max_wait:
            raise ValueError("Could not get transaction receipt")
        time.sleep(1)
    return txn_receipt


def wait_for_block(rpc_client, block_number, max_wait=60):
    start = time.time()
    while time.time() < start + max_wait:
        if rpc_client.get_block_number() >= block_number:
            break
        time.sleep(1)
    else:
        raise ValueError("Did not reach block")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

DIR = os.path.dirname(os.path.abspath(__file__))

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

version = '0.5.0-rc1'

readme = open(os.path.join(DIR, 'README.md')).read()


setup(
    name='populus',
    version=version,
    description="""Ethereum Development Framework""",
    long_description=readme,
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    url='https://github.com/pipermerriam/populus',
    include_package_data=True,
    py_modules=['populus'],
    install_requires=[
        "click>=5.0",
        # Until https://github.com/ethereum/ethash/issues/72 is resolved
        # "ethash>=23.1",
        "ethereum>=0.9.73",
        "requests>=2.7.0",
        "eth-testrpc>=0.1.16",
        "pytest>=2.7.2",
        "ethereum-rpc-client>=0.2.2",
        "watchdog>=0.8.3",
        "Flask>=0.10.1",
    ],
    dependency_links=[
        # Until https://github.com/ethereum/ethash/issues/72 is resolved
        "https://github.com/ethereum/ethash/archive/v23.1.tar.gz",
    ],
    license="MIT",
    zip_safe=False,
    entry_points={
        'console_scripts': ["populus=populus.cli:main"],
        'pytest11': ['ethereum=populus.plugin'],
    },
    keywords='ethereum pytest',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



import subprocess
import itertools
import functools
import yaml
import re

from populus import utils


class CompileError(Exception):
    pass


SOLC_BINARY = 'solc'


version_regex = re.compile('Version: ([0-9]+\.[0-9]+\.[0-9]+(-[a-f0-9]+)?)')


is_solc_available = functools.partial(utils.is_executable_available, 'solc')


def solc_version():
    version_string = subprocess.check_output(['solc', '--version'])
    version = version_regex.search(version_string).groups()[0]
    return version


def solc(source=None, input_files=None, add_std=True,
         combined_json='json-abi,binary,sol-abi,natspec-dev,natspec-user',
         raw=False, rich=True):

    if source and input_files:
        raise ValueError("`source` and `input_files` are mutually exclusive")
    elif source is None and input_files is None:
        raise ValueError("Must provide either `source` or `input_files`")

    command = ['solc']
    if add_std:
        command.append('--add-std=1')

    if combined_json:
        command.extend(('--combined-json', combined_json))

    if input_files:
        command.extend(itertools.chain(*zip(itertools.repeat('--input-file'), input_files)))

    p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    if source:
        stdoutdata, stderrdata = p.communicate(input=source)
    else:
        stdoutdata, stderrdata = p.communicate()

    if p.returncode:
        raise CompileError('compilation failed')

    if raw:
        return stdoutdata

    contracts = yaml.safe_load(stdoutdata)['contracts']

    for contract_name, data in contracts.items():
        data['json-abi'] = yaml.safe_load(data['json-abi'])
        data['sol-abi'] = yaml.safe_load(data['sol-abi'])
        data['natspec-dev'] = yaml.safe_load(data['natspec-dev'])
        data['natspec-user'] = yaml.safe_load(data['natspec-user'])

    sorted_contracts = sorted(contracts.items(), key=lambda c: c[0])

    if not rich:
        return sorted_contracts

    compiler_version = solc_version()

    return {
        contract_name: {
            'code': "0x" + contract.get('binary'),
            'info': {
                'abiDefinition': contract.get('json-abi'),
                'compilerVersion': compiler_version,
                'developerDoc': contract.get('natspec-dev'),
                'language': 'Solidity',
                'languageVersion': '0',
                'source': source,  # what to do for files?
                'userDoc': contract.get('natspec-user')
            },
        }
        for contract_name, contract
        in sorted_contracts
    }
